(* somme des éléments de l *)
let rec somme (l: int list) : int = 
	match l with 
	| [] -> 0
	| x :: q -> x + somme q 

(* true si x est dans l, false sinon *)
let rec recherche (x: 'a) (l: 'a list) : bool =
	match l with 
	| [] -> true
	(* On ne peut pas mettre x dans le motif
	   car un motif sert à décrire une forme
	   et pas une valeur *)
	| y :: q -> if x = y then true else recherche x q


(* [x1; .. xn; y1; ...; ym],
   où l1 = [x1; ...; xn]
   et l2 = [y1; ...; ym] *)

 let rec concatener (l1: 'a list) (l2: 'a list) : 'a list =
 	match l1 with 
 	| [] -> l2
 	| x :: q -> x :: concatener q l2

(* concaténation de toutes les listes contenues dans l *)
 let rec multi_concat (ll: 'a list list) : 'a list =
 	match ll with 
 	| [] -> []
 	| l :: qq -> concatener l (multi_concat qq)

 (* [f x1; f x2; ...; f xn],
    où l = [x1; x2; ...; xn] *)
let rec map (f: 'a -> 'b) (l: 'a list) : 'b list =
	match l with 
	| [] -> []
	| x :: q -> f x :: map f q

(* concaténation de (l1 renversée) et l2 *)
let rec rev_concat (l1: 'a list) (l2: 'a list) : 'a list =
	match l1 with 
	| [] -> l2
	| x :: q -> rev_concat q (x::l2)

(* l dans l'ordre inverse *)
let reverse (l: 'a list) : 'a list =
	rev_concat l []

(* [0; ...; n-1] concaténé avec l *)
let rec range_concat (n: int) (l: int list) : int list =
	if n = 0 then l
	else range_concat (n-1) ((n-1)::l)

(* [0; ...; n-1] *)
let range n = range_concat n []

(* [f 0; ...; f (n-1)] *)
let make_list f n =
	map f (range n)

(*** compose_liste ***)
(* Il y a différentes manières de coder cette fonction. 
 * La plus simple est de littéralement faire une fonction
 * qui prend en argument une liste de fonctions et qui 
 * renvoie la fonction obtenue en les composant toutes.
 * Autrement dit, pour l une liste de fonction, on décrit
 * directement ce qu'est la fonction compose_liste l
 *
 *
 * Une deuxième manière, un peu plus technique, est de
 * directement décrire ce qu'est (compose_list l) x pour
 * un x quelconque. Sur un exemple plus simple, c'est comme
 * si on avait pris la fonction
 * let compose f g =
 *	  fun x -> f (g x)
 * qui renvoie la composée de f g
 * et qu'on l'avait plutôt définie ainsi:
 * let compose f g x =
 *		f (g x)
 *)


 (* Première méthode *)
 let rec compose_list (lf: ('a -> 'a) list) : 'a ->'a =
 	match lf with 
 	| [] -> fun x -> x 
 	| f :: qf -> fun x -> f (compose_list qf x)

(* Deuxième méthode *)
let rec compose_list2 (lf: ('a -> 'a) list) (x: 'a) : 'a =
	match lf with 
	| [] -> x (* si on applique une liste vide de fonctions sur x on obtient x *)
	| f :: qf -> f (compose_list qf x)

(* Question 11 *)
let petit_f k x = 
	if x mod (2+k) = 0 then x / (2+k)
	else 3*x + k


let grand_f_n n=
	compose_list (make_list petit_f n)

let grand_f n =
	let ff = grand_f_n n in
	somme (make_list (fun i -> ff i) n)

let _ = 
	assert (grand_f 3 = 83);
	assert (grand_f 20 = 60321545051);
	assert (grand_f 30 = 12254504478085045)
