let rec taille (l: int list) : int = 
	match l with 
	| [] -> 0
	| x :: q -> 1 + taille q 

let rec recherche (a: 'a) (l: 'a list) : bool =
	match l with 
	| [] -> true
	| x :: q -> if x = a then true else recherche x q

let rec somme (l: int list) : int = 
	match l with 
	| [] -> 0
	| x :: q -> x + somme q 

let rec multi_concat (ll: 'a list list) : 'a list =
 	match ll with 
 	| [] -> []
 	| l :: qq -> List.append l (multi_concat qq) (* List.append fait une concaténation *)

let rec string_cat (l: string list) : string = 
	match l with 
	| [] -> ""
	| x :: q -> x ^ string_cat q 

let rec fold (f: 'a -> 'b -> 'b) (l: 'a list) (a: 'b) =
	match l with 
	| [] -> a
	| x :: q -> f x (fold f q a)


(* nouvelles définitions de taille, somme, 
   recherche, multi_concat et string_cat 
   en utilisant fold.
   On remarque que systématiquement, 
   le cas | [] -> se retrouve dans le dernier argument
   et le cas | x :: q -> se retrouve dans le premier argument *)

let taille2 l = 
	(* si on a une liste de la forme x :: q et qu'on
	   a déjà calculé la taille de q, que l'on note r,
	   alors la taille de l est 1 + r *)
	fold (fun x r -> 1 + r) l 0

let somme2 l = 
	(* si on a une liste de la forme x :: q et qu'on
	   a déjà calculé la somme de q, que l'on note r,
	   alors la somme de l est x + r *)
	fold (fun x r -> x + r) l 0 (*ou bien fold (+) l 0 *)

let recherche2 a l = 
	(* si on a une liste de la forme x :: q et qu'on
	   a déjà cherché a dans q, en notant r le résultat,
	   alors a est dans l si x = a ou bien si a est dans
	   q, i.e. si x = a ou sinon si r est vrai *)
	fold (fun x r -> if x = a then true else r) l false

let multi_concat2 ll =
	(* même raisonnement *)
	fold (fun l r -> List.append l r) ll []

let string_cat2 l =
	fold (fun x r -> x^r) l "" (* ou bien fold (^) l "" *)

let compose_list2 lf =
	(* si on a déjà calculé r la composition de qf
	   et qu'on veut calculer la composition de f :: qf
	   il faut faire la fonction (fun x -> f (r x)),
	   ce qu'on peut exprimer ainsi: *)
	fold (fun f r -> (fun x -> f (r x))) lf (fun x -> x)
	(* ou bien fold (fun f r x -> f (r x) ) lf (fun x -> x) *)

(* liste des éléments de l qui vérifient prop *)
let rec filter1 (prop: 'a -> bool) (l: 'a list) : 'a list =
	match l with
	| [] -> []
	| x :: q -> if prop x then x :: filter1 prop q 
	            else             filter1 prop q 

let filter prop l =
	fold (fun x r -> if prop x then x::r else r) l []

let map f l =
	fold (fun x r -> f x :: r) l []

(* liste des caractères de s à partir de l'indice
   i inclus *)
let rec list_of_string_from_i (s: string) (i: int) : char list =
	if i = String.length s then []
	else s.[i] :: list_of_string_from_i s (i+1)

(* liste des caractères de s *)
let list_of_string (s: string) : char list =
	list_of_string_from_i s 0

let string_of_list (l: char list) : string =
	String.of_seq (List.to_seq l)

(* liste de strings obtenus en séparant s[i..] selon
   le séparateur sep, avec curr qui contient les
   caractères lus dans le segment actuel. *)
let rec split_from_i 
	(s: string) (sep: char) (i: int) (curr: char list) : string list =
	if i = String.length s then 
		(* curr est stockée à l'envers, il faut la renverser *)
		[string_of_list (List.rev curr)]
	else if s.[i] = sep then (* on a fini de lire un segment *)
		string_of_list (List.rev curr) :: split_from_i s sep (i+1) []
    else (* on rajoute la lettre au segment actuel *)
        split_from_i s sep (i+1) (s.[i] :: curr)

let split (s: string) (sep: char) : string list =
	split_from_i s sep 0 []