type couleur = Coeur | Pique | Carreau | Trefle

let couleur1 = Pique

(* true si c est une couleur rouge, faux sinon *)
let est_rouge (c: couleur) : bool = match c with
	| Coeur   -> true
	| Carreau -> true
	| Pique   -> false
	| Trefle  -> false

let _ =
	assert (est_rouge Carreau) (* vaut true *)

type carte = 
	| Nombre of (int * couleur) (* Cartes 2 à 10 *)
	| Valet of couleur
	| Dame of couleur
	| Roi of couleur
	| As of couleur

let carte_1 = Valet Coeur 
let carte_2 = Nombre (9, Pique)

(* Renvoie la couleur de ca *)
let couleur_de_carte (ca: carte) : couleur = 
	match ca with
	| Nombre (n, cou) -> cou 
	| Valet cou -> cou
	| Dame cou -> cou
	| Roi cou -> cou
	| As cou -> cou

let _ =
	assert (couleur_de_carte (Roi Trefle) = Trefle)

(* chaine de caractère représentant c *)
let string_of_couleur (c: couleur) : string = 
	match c with
	| Coeur -> "coeur"
	| Pique -> "pique"
	| Trefle -> "trèfle"
	| Carreau -> "carreau"

(* chaine de caractère représentant ca *)
let string_of_carte (ca: carte) : string = 
	match ca with
	| Nombre (n, cou) -> string_of_int n ^ " de " ^ string_of_couleur cou
	| Valet cou -> "Valet de " ^ string_of_couleur cou
	| Dame cou -> "Dame de " ^ string_of_couleur cou
	| Roi cou -> "Roi de " ^ string_of_couleur cou
	| As cou -> "As de " ^ string_of_couleur cou

(* Insère la carte `ca` dans la main `main` selon
   l'ordre naturel d'OCaml. 
   Précondition: main est triée *)
let rec inserer1 (ca: carte) (main: carte list) : carte list = 
	match main with
	| [] -> [ca]
	| x :: q -> if ca < x then ca::x::q 
	          else x :: inserer1 ca q

(* Renvoie -1, 0 ou 1 selon si ca1 est inférieure, égale
   ou supérieure à ca2 dans l'ordre suivant:
   coeur < carreau < pique < trefle et au sein d'une
   couleur 2 < 3 < ... < 10 < valet < dame < roi < as *)
let comparer_cartes (ca1: carte) (ca2: carte) : int = 
	if ca1 = ca2 then 0 else
	let cou1, cou2 = couleur_de_carte ca1, couleur_de_carte ca2 in
	if cou1 < cou2 then -1
	else if cou1 > cou2 then 1
	else if ca1 < ca2 then -1
	else                    1

(* Insère ca dans main selon l'ordre dicté par comparer_cartes.
   Précondition: main est triée selon ce même ordre *)
let rec inserer2 (ca: carte) (main: carte list) = 
	match main with
	| [] -> [ca]
	| x :: q -> if comparer_cartes ca x <= 0 then ca::x::q 
	          else x :: inserer2 ca q

(* liste triée selon comparer_cartes contenant les éléments de l *)
let rec tri_insertion (l: carte list) : carte list = 
	match l with
	| [] -> []
	| x::q -> inserer2 x (tri_insertion q);;

(** Quelques opérateurs et fonctions pratiques sur les listes **)
(* @ fait une concaténation. ATTENTION c'est un opérateur
   très dangereux car il donne l'impression d'être aussi simple
   qu'une addition, mais en réalité il s'exécute comme la fonction 
   concatener de l'exercice 1, avec une complexité linéaire en la
   taille de la liste de gauche. Il faut s'en servir avec parcimonie.
   Ici c'est justifiable car les listes que l'on concatène sont petites. *)

(* Liste contenant toutes les cartes de la couleur cou *)
let gen_couleur (cou: couleur) : carte list = 
	let rec nombres cou k = match k with
		|  1 -> []
		|  _ -> Nombre (k, cou):: nombres cou (k-1)
	in nombres cou 10 @ ([Valet cou; Dame cou; Roi cou; As cou])

(* List.concat fait la même chose que multiconcat de l'exercice 1:
   elle prend une liste de listes et les concatènes toutes. *)
(* List.map est la fonction map des exercices 1 et 2 *)

(* renvoie un deck entier trié *)
let gen_deck () = 
	(* on génère la liste de chaque couleur et on concatène le tout *)
	let deck = List.concat (List.map gen_couleur [Coeur; Pique; Trefle; Carreau])
	in tri_insertion deck







